@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use base qw( Exporter );
 
-our $VERSION = '0.16';
+our $VERSION = '0.18';
 
 use AnyEvent::RipeRedis;
 use AnyEvent::RipeRedis::Error;
@@ -15,7 +15,7 @@ use List::MoreUtils qw( bsearch );
 use Scalar::Util qw( looks_like_number weaken );
 use Carp qw( croak );
 
-our %ERROR_CODES;
+my %ERROR_CODES;
 
 BEGIN {
   %ERROR_CODES = %AnyEvent::RipeRedis::Error::ERROR_CODES;
@@ -75,13 +75,8 @@ my %PREDEFINED_CMDS = (
   sort        => { readonly => 0, key_pos => 1 },
   zunionstore => { readonly => 0, key_pos => 1 },
   zinterstore => { readonly => 0, key_pos => 1 },
-  eval        => { readonly => 0, key_pos => 3 },
-  evalsha     => { readonly => 0, key_pos => 3 },
-  multi       => { readonly => 0, key_pos => 1, force_slot => 1 },
-  watch       => { readonly => 0, key_pos => 1, force_slot => 1 },
-  exec        => { readonly => 0, key_pos => 0, unforce_slot => 1 },
-  discard     => { readonly => 0, key_pos => 0, unforce_slot => 1 },
-  unwatch     => { readonly => 0, key_pos => 0, unforce_slot => 1 },
+  eval        => { readonly => 0, movablekeys => 1, key_pos => 0 },
+  evalsha     => { readonly => 0, movablekeys => 1, key_pos => 0 },
 );
 
 my %SUB_CMDS = (
@@ -160,8 +155,8 @@ sub disconnect {
 }
 
 sub nodes {
-  my $self = shift;
-  my $key  = shift;
+  my $self         = shift;
+  my $key          = shift;
   my $allow_slaves = shift;
 
   return unless defined $self->{_nodes_pool};
@@ -171,9 +166,11 @@ sub nodes {
     $slot = hash_slot($key);
   }
 
-  my @nodes = $self->_nodes( $slot, $allow_slaves );
+  my $nodes = $self->_nodes( $slot, $allow_slaves );
 
-  return @{ $self->{_nodes_pool} }{@nodes};
+  return wantarray
+      ? @{ $self->{_nodes_pool} }{ @{$nodes} }
+      : $self->{_nodes_pool}{ $nodes->[0] };
 }
 
 sub refresh_interval {
@@ -235,31 +232,49 @@ sub crc16 {
 sub hash_slot {
   my $key = shift;
 
-  my $tag = $key;
+  my $hashtag = $key;
 
   if ( $key =~ m/\{([^}]*?)\}/ ) {
     if ( length $1 > 0 ) {
-      $tag = $1;
+      $hashtag = $1;
     }
   }
 
-  return crc16($tag) % MAX_SLOTS;
+  return crc16($hashtag) % MAX_SLOTS;
 }
 
 sub _init {
-  my $self  = shift;
+  my $self = shift;
 
   $self->{_init_state} = S_IN_PROGRESS;
   undef $self->{_refresh_timer};
 
+  unless ( defined $self->{_nodes_pool} ) {
+    my %nodes_pool;
+
+    foreach my $node_params ( @{ $self->{startup_nodes} } ) {
+      my $hostport = "$node_params->{host}:$node_params->{port}";
+
+      unless ( defined $nodes_pool{$hostport} ) {
+        $nodes_pool{$hostport} = $self->_new_node(
+            $node_params->{host}, $node_params->{port}, 1 );
+      }
+    }
+
+    $self->{_nodes_pool} = \%nodes_pool;
+    $self->{_nodes}      = [ keys %nodes_pool ];
+  }
+
   weaken($self);
 
-  $self->_discover_nodes(
+  $self->_discover_cluster(
     sub {
       my $err = $_[1];
 
       if ( defined $err ) {
         $self->{_init_state} = S_NEED_DO;
+
+        $self->{_ready} = 0;
         $self->_abort($err);
 
         return;
@@ -285,63 +300,60 @@ sub _init {
   return;
 }
 
-sub _discover_nodes {
+sub _discover_cluster {
   my $self = shift;
   my $cb   = shift;
 
-  my $nodes_pool = $self->{_nodes_pool};
-
-  my @nodes;
-  if ( %{$nodes_pool} ) {
-    @nodes = keys %{$nodes_pool};
-  }
-  else {
-    foreach my $node_params ( @{ $self->{startup_nodes} } ) {
-      my $hostport = "$node_params->{host}:$node_params->{port}";
-
-      unless ( defined $nodes_pool->{$hostport} ) {
-        $nodes_pool->{$hostport} = $self->_new_node(
-            $node_params->{host}, $node_params->{port}, 1 );
-      }
-
-      push( @nodes, $hostport );
-    }
-  }
-
   weaken($self);
 
   $self->_execute(
-    { name => 'cluster_slots',
+    { name => 'cluster_state',
       args => [],
 
       on_reply => sub {
-        my $slots = shift;
-        my $err   = shift;
+        my $err = $_[1];
 
         if ( defined $err ) {
           $cb->( undef, $err );
           return;
         }
 
-        $self->_prepare_nodes_pool( $slots,
-          sub {
-            unless ( defined $self->{_commands} ) {
-              $self->_discover_commands($cb);
-              return;
-            }
+        $self->_execute(
+          { name => 'cluster_slots',
+            args => [],
 
-            $cb->();
-          }
+            on_reply => sub {
+              my $slots = shift;
+              my $err   = shift;
+
+              if ( defined $err ) {
+                $cb->( undef, $err );
+                return;
+              }
+
+              $self->_prepare_nodes( $slots,
+                sub {
+                  unless ( defined $self->{_commands} ) {
+                    $self->_load_commands($cb);
+                    return;
+                  }
+
+                  $cb->();
+                }
+              );
+            }
+          },
+          $self->{_nodes}
         );
       }
     },
-    @nodes,
+    $self->{_nodes}
   );
 
   return;
 }
 
-sub _prepare_nodes_pool {
+sub _prepare_nodes {
   my $self      = shift;
   my $slots_raw = shift;
   my $cb        = shift;
@@ -351,7 +363,7 @@ sub _prepare_nodes_pool {
   my @masters;
   my @slaves;
 
-  my $old_nodes_pool = $self->{_nodes_pool};
+  my $nodes_pool_old = $self->{_nodes_pool};
 
   foreach my $range ( @{$slots_raw} ) {
     my $range_start = shift @{$range};
@@ -364,8 +376,8 @@ sub _prepare_nodes_pool {
       my $hostport = "$node_info->[0]:$node_info->[1]";
 
       unless ( defined $nodes_pool{$hostport} ) {
-        if ( defined $old_nodes_pool->{$hostport} ) {
-          $nodes_pool{$hostport} = $old_nodes_pool->{$hostport};
+        if ( defined $nodes_pool_old->{$hostport} ) {
+          $nodes_pool{$hostport} = delete $nodes_pool_old->{$hostport};
         }
         else {
           $nodes_pool{$hostport} = $self->_new_node( @{$node_info}[ 0, 1 ] );
@@ -389,7 +401,12 @@ sub _prepare_nodes_pool {
 
   @slots = sort { $a->[0] <=> $b->[0] } @slots;
 
+  foreach my $node ( values %{$nodes_pool_old} ) {
+    $node->disconnect;
+  }
+
   $self->{_nodes_pool} = \%nodes_pool;
+  $self->{_nodes}      = [ keys %nodes_pool ];
   $self->{_masters}    = \@masters;
   $self->{_slots}      = \@slots;
 
@@ -421,18 +438,17 @@ sub _prepare_slaves {
   };
 
   foreach my $hostport ( @{$slaves} ) {
-    $self->_execute( $cmd, $hostport );
+    $self->_execute( $cmd, [ $hostport ] );
   }
 
   return;
 }
 
-sub _discover_commands {
+sub _load_commands {
   my $self = shift;
   my $cb   = shift;
 
   weaken($self);
-  my @nodes = keys %{ $self->{_nodes_pool} };
 
   $self->_execute(
     { name => 'command',
@@ -473,7 +489,7 @@ sub _discover_commands {
         $cb->();
       }
     },
-    @nodes,
+    $self->{_nodes}
   );
 
   return;
@@ -593,97 +609,90 @@ sub _route {
   my $self = shift;
   my $cmd  = shift;
 
-  if ( $cmd->{name} eq 'multi'
-    && !defined $cmd->{args}[0] )
-  {
-    croak 'Hash tag for "multi" command not specified';
-  }
-
   unless ( $self->{_ready} ) {
     if ( $self->{_init_state} == S_NEED_DO ) {
       $self->_init;
     }
-
     push( @{ $self->{_input_queue} }, $cmd );
 
     return;
   }
 
+  my $key;
   my $cmd_info = $self->{_commands}{ $cmd->{kwds}[0] };
 
-  my $slot;
-  my $allow_slaves;
+  if ( defined $cmd_info ) {
+    my @tokens = ( @{ $cmd->{kwds} }, @{ $cmd->{args} } );
 
-  if ( defined $self->{_forced_slot} ) {
-    $slot = $self->{_forced_slot};
-    $cmd->{is_forced_slot} = 1;
-
-    if ( defined $cmd_info && $cmd_info->{unforce_slot} ) {
-      undef $self->{_forced_slot};
-    }
-  }
-  else {
-    my $key;
-
-    if ( defined $cmd_info ) {
-      my @tokens = ( @{ $cmd->{kwds} }, @{ $cmd->{args} } );
+    if ( $cmd_info->{key_pos} > 0 ) {
       $key = $tokens[ $cmd_info->{key_pos} ];
     }
-
-    $allow_slaves = $self->{allow_slaves};
-
-    if ( defined $key ) {
-      $slot = hash_slot($key);
-
-      if ( $cmd_info->{force_slot}
-        && !defined $self->{_forced_slot} )
-      {
-        $self->{_forced_slot} = $slot;
-      }
-
-      $allow_slaves &&= $cmd_info->{readonly};
+    # Exception for EVAL and EVALSHA commands
+    elsif ( $cmd_info->{movablekeys}
+      && $tokens[2] > 0 )
+    {
+      $key = $tokens[3];
     }
   }
 
-  if ( $cmd->{name} eq 'multi' ) {
-    $cmd->{args} = [];
+  my $slot;
+  my $allow_slaves = $self->{allow_slaves};
+
+  if ( defined $key ) {
+    $slot = hash_slot($key);
+    $allow_slaves &&= $cmd_info->{readonly};
   }
 
-  my @nodes = $self->_nodes( $slot, $allow_slaves );
-  $self->_execute( $cmd, @nodes );
+  my $nodes = $self->_nodes( $slot, $allow_slaves );
+  $self->_execute( $cmd, $nodes );
 
   return;
 }
 
 sub _execute {
-  my $self  = shift;
-  my $cmd   = shift;
-  my @nodes = @_;
+  my $self       = shift;
+  my $cmd        = shift;
+  my $nodes      = shift;
+  my $node_index = shift;
+  my $fails_cnt  = shift || 0;
 
-  my $hostport;
-  if ( scalar @nodes > 1 ) {
-    my $rand_index = int( rand( scalar @nodes ) );
-    $hostport = splice( @nodes, $rand_index, 1 );
+  unless ( defined $node_index ) {
+    $node_index = int( rand( scalar @{$nodes} ) );
   }
-  else {
-    $hostport = shift @nodes;
+  elsif ( $node_index == scalar @{$nodes} ) {
+    $node_index = 0;
   }
-  my $node = $self->{_nodes_pool}{$hostport};
+  my $hostport = $nodes->[$node_index];
+  my $node     = $self->{_nodes_pool}{$hostport};
+
+  my $cmd_name = $cmd->{name} eq 'cluster_state'
+      ? 'cluster_info'
+      : $cmd->{name};
 
   weaken($self);
 
-  $node->execute( $cmd->{name}, @{ $cmd->{args} },
+  $node->execute( $cmd_name, @{ $cmd->{args} },
     { on_reply => sub {
         my $reply = shift;
         my $err   = shift;
 
+        if ( $cmd->{name} eq 'cluster_state' ) {
+          unless ( defined $err ) {
+            if ( $reply->{cluster_state} eq 'ok' ) {
+              $reply = 1;
+            }
+            else {
+              $err = _new_error( 'CLUSTERDOWN The cluster is down',
+                  E_CLUSTER_DOWN );
+            }
+          }
+        }
+
         if ( defined $err ) {
           my $err_code   = $err->code;
           my $nodes_pool = $self->{_nodes_pool};
 
-          if ( ( $err_code == E_MOVED || $err_code == E_ASK )
-            && !$cmd->{is_forced_slot} )
-          {
+          if ( $err_code == E_MOVED || $err_code == E_ASK ) {
             if ( $err_code == E_MOVED ) {
               $self->{_init_state} = S_NEED_DO;
               $self->{_ready}      = 0;
@@ -696,7 +705,7 @@ sub _execute {
               $nodes_pool->{$fwd_hostport} = $self->_new_node( $host, $port );
             }
 
-            $self->_execute( $cmd, $fwd_hostport );
+            $self->_execute( $cmd, [ $fwd_hostport ] );
 
             return;
           }
@@ -707,8 +716,10 @@ sub _execute {
             $on_node_error->( $err, $node->host, $node->port );
           }
 
-          if ( $err_code != E_CONN_CLOSED_BY_CLIENT && @nodes ) {
-            $self->_execute( $cmd, @nodes );
+          if ( $err_code != E_CONN_CLOSED_BY_CLIENT
+            && ++$fails_cnt < scalar @{$nodes} )
+          {
+            $self->_execute( $cmd, $nodes, ++$node_index, $fails_cnt );
             return;
           }
 
@@ -741,13 +752,13 @@ sub _nodes {
     @{ $self->{_slots} };
 
     return $allow_slaves
-        ? @{ $range->[2] }
-        : ( $range->[2][0] );
+        ? $range->[2]
+        : [ $range->[2][0] ];
   }
 
   return $allow_slaves
-      ? keys %{ $self->{_nodes_pool} }
-      : @{ $self->{_masters} };
+      ? $self->{_nodes}
+      : $self->{_masters};
 }
 
 sub _process_input_queue {
@@ -766,14 +777,15 @@ sub _process_input_queue {
 sub _reset_internals {
   my $self = shift;
 
-  $self->{_nodes_pool}    = {};
+  $self->{_nodes_pool}    = undef;
+  $self->{_startup_nodes} = undef;
+  $self->{_nodes}         = undef;
   $self->{_masters}       = undef;
   $self->{_slots}         = undef;
   $self->{_commands}      = undef;
   $self->{_init_state}    = S_NEED_DO;
   $self->{_refresh_timer} = undef;
   $self->{_ready}         = 0;
-  $self->{_forced_slot}   = undef;
 
   return;
 }
@@ -875,6 +887,7 @@ AnyEvent::RipeRedis::Cluster - Non-blocking Redis Cluster client
     startup_nodes => [
       { host => 'localhost', port => 7000 },
       { host => 'localhost', port => 7001 },
+      { host => 'localhost', port => 7002 },
     ],
   );
 
@@ -941,6 +954,7 @@ L<http://redis.io/topics/cluster-spec>
     startup_nodes => [
       { host => 'localhost', port => 7000 },
       { host => 'localhost', port => 7001 },
+      { host => 'localhost', port => 7002 },
     ],
     connection_timeout => 5,
     read_timeout       => 5,
@@ -1228,15 +1242,16 @@ convenient.
 
 =head1 TRANSACTIONS
 
-The client sends all commands of the transaction to the one master node,
-which will be selected by the hash tag in C<MULTI> command or by the first key
-in the C<WATCH> command. The hash tag of the C<MULTI> command is an extension
-of the client and it is not transferred to the Redis server.
+To perform the transaction you must get the master node by the key using
+C<nodes> method and then execute all commands on this node. Nodes must be
+discovered first.
 
-  $cluster->multi('foo');
-  $cluster->set( '{foo}bar', "some\r\nstring" );
-  $cluster->set( '{foo}car', 42 );
-  $cluster->exec(
+  $node = $cluster->nodes('foo');
+
+  $node->multi;
+  $node->set( '{foo}bar', "some\r\nstring" );
+  $node->set( '{foo}car', 42 );
+  $node->exec(
     sub {
       my $reply = shift;
       my $err   = shift;
@@ -1282,7 +1297,8 @@ aborted.
 
 =head2 nodes( [ $key ] [, $allow_slaves ] )
 
-Gets particular nodes of the cluster. Nodes must be discovered first.
+Gets particular nodes of the cluster. Nodes must be discovered first. In scalar
+context method returns the first node from the list.
 
 Getting all master nodes of the cluster:
 
@@ -1292,9 +1308,9 @@ Getting all nodes of the cluster, including slave nodes:
 
   my @nodes = $cluster->nodes( undef, 1 );
 
-Getting master nodes by the key:
+Getting master node by the key:
 
-  my @master_nodes = $cluster->nodes('foo');
+  my $master_node = $cluster->nodes('foo');
 
 Getting nodes by the key, including slave nodes:
 
